namespace SharpScript
{
	#region Using Directives

	using System;
	using System.Collections.Generic;
	using System.Diagnostics;
	using System.IO;
	using System.Linq;
	using System.Reflection;
	using System.Text;
	using System.Threading.Tasks;
	using Menees;
	using Microsoft.CodeAnalysis;
	using Microsoft.CodeAnalysis.CSharp;
	using Microsoft.CodeAnalysis.Emit;
	using Microsoft.CodeAnalysis.Text;
	using Microsoft.CodeAnalysis.VisualBasic;

	#endregion

	internal sealed class RoslynCompiler : ScriptCompiler
	{
		#region Constructors

		public RoslynCompiler(
			ScriptTypeProvider stp,
			ScriptDirectives directives,
			ScriptParameters parameters,
			Task<IEnumerable<string>> assemblyFoldersTask)
			: base(stp, directives, parameters, assemblyFoldersTask)
		{
		}

		#endregion

		#region Public Methods

		public override Assembly Compile(bool throwOnError)
		{
			IEnumerable<MetadataReference> references = this.CreateReferences(fileName => MetadataReference.CreateFromFile(fileName));

			string scriptFileName = this.Parameters.FileName;
			SourceText sourceText;
			using (Stream sourceTextStream = File.OpenRead(scriptFileName))
			{
				sourceText = SourceText.From(sourceTextStream);
			}

			// Compile the script.
			Compilation compilation;
			if (this.TypeProvider.ScriptType == ScriptType.VB)
			{
				compilation = VBCompiler.CreateCompilation(this.Parameters, sourceText, references);
			}
			else
			{
				compilation = CSharpCompiler.CreateCompilation(this.Parameters, sourceText, references);
			}

			// Add any output files required by script directives like #com_reference.
			Assembly result = null;
			if (this.Parameters.Debug)
			{
				// Append new extensions instead of changing the script extension because in the Samples
				// directory we can have multiple scripts with the same name but different extensions, and
				// we want to be able to run them at the same time.
				string exeName = scriptFileName + ".exe";
				string pdbName = scriptFileName + ".pdb";
				EmitResult emit = compilation.Emit(exeName, pdbName);
				this.AddOutputFile(exeName);
				this.AddOutputFile(pdbName);
				if (this.VerifyEmit(emit, sourceText, throwOnError))
				{
					AssemblyName assemblyName = AssemblyName.GetAssemblyName(exeName);
					result = Assembly.Load(assemblyName);
				}
			}
			else
			{
				using (MemoryStream memory = new MemoryStream())
				{
					EmitResult emit = compilation.Emit(memory);
					if (this.VerifyEmit(emit, sourceText, throwOnError))
					{
						memory.Seek(0, SeekOrigin.Begin);
						result = Assembly.Load(memory.ToArray());
					}
				}
			}

			return result;
		}

		#endregion

		#region Private Methods

		private static T AddCommonOptions<T>(T customOptions, ScriptParameters parameters)
			where T : CompilationOptions
		{
			// We have to specify a custom assemblyIdentityComparer to correctly handle references
			// to multiple mscorlib assemblies (e.g., if a referenced assembly was compiled against 2.0).
			// See https://github.com/ctaggart/froto/issues/2 and https://gist.github.com/m0sa/d0f5eab1eed43952623f.
			bool debug = parameters.Debug;
			CompilationOptions result = customOptions
				.WithOutputKind(parameters.InConsole ? OutputKind.ConsoleApplication : OutputKind.WindowsApplication)
				.WithOptimizationLevel(debug ? OptimizationLevel.Debug : OptimizationLevel.Release)
				.WithGeneralDiagnosticOption(ReportDiagnostic.Error)
				.WithAssemblyIdentityComparer(DesktopAssemblyIdentityComparer.Default)
				.WithStrongNameProvider(new DesktopStrongNameProvider());

			return (T)result;
		}

		private bool VerifyEmit(EmitResult emit, SourceText sourceText, bool throwOnError)
		{
			if (!emit.Success)
			{
				const int BufferSize = 1024;
				StringBuilder sb = new StringBuilder(BufferSize);
				sb.Append(Properties.Resources.CompileErrors);
				sb.Append(":\r\n");

				IEnumerable<Diagnostic> errors = emit.Diagnostics.Where(diagnostic =>
					diagnostic.IsWarningAsError ||
					diagnostic.Severity == DiagnosticSeverity.Error);
				foreach (Diagnostic error in errors)
				{
					string sourceCodeLine = null;
					Location location = error.Location;
					if (location.IsInSource)
					{
						FileLinePositionSpan lineSpan = location.GetLineSpan();
						LinePosition linePosition = lineSpan.StartLinePosition;
						sb.AppendFormat(Properties.Resources.LineCharacter, linePosition.Line + 1, linePosition.Character + 1);
						sourceCodeLine = sourceText.Lines[linePosition.Line].ToString();
					}

					// Show the standard error information.
					string type = error.Severity == DiagnosticSeverity.Warning ? Properties.Resources.Warning : Properties.Resources.Error;
					sb.AppendFormat("{0} {1}: {2}\r\n", type, error.Id, error.GetMessage());

					if (!string.IsNullOrWhiteSpace(sourceCodeLine))
					{
						sb.Append('\t');
						sb.Append(sourceCodeLine.Trim());
						sb.Append("\r\n");
					}

					sb.Append("\r\n");
				}

				this.FailCompile(sb.ToString(), throwOnError);
			}

			return emit.Success;
		}

		#endregion

		#region Private Types

		private static class CSharpCompiler
		{
			#region Public Methods

			public static Compilation CreateCompilation(
				ScriptParameters parameters,
				SourceText sourceText,
				IEnumerable<MetadataReference> references)
			{
				SyntaxTree syntaxTree = ParseScript(parameters, sourceText);
				CSharpCompilationOptions options = CreateCompilationOptions(parameters);
				CSharpCompilation result = CSharpCompilation.Create(parameters.ScriptName, new[] { syntaxTree }, references, options);
				return result;
			}

			#endregion

			#region Private Methods

			private static SyntaxTree ParseScript(ScriptParameters parameters, SourceText sourceText)
			{
				CSharpParseOptions parseOptions = new CSharpParseOptions(
					documentationMode: DocumentationMode.None,
					kind: SourceCodeKind.Regular,
					preprocessorSymbols: parameters.Debug ? new[] { "TRACE", "DEBUG" } : new[] { "TRACE" });

				// We must tell the parser the FileName that the sourceText came from in order for debugging
				// to work (e.g., to have file and line info in the .pdb).
				SyntaxTree result = CSharpSyntaxTree.ParseText(sourceText, parseOptions, parameters.FileName);
				return result;
			}

			private static CSharpCompilationOptions CreateCompilationOptions(ScriptParameters parameters)
			{
				CSharpCompilationOptions result = new CSharpCompilationOptions(
					OutputKind.WindowsApplication,
					checkOverflow: parameters.Debug);

				result = AddCommonOptions(result, parameters);
				return result;
			}

			#endregion
		}

		private static class VBCompiler
		{
			#region Public Methods

			public static Compilation CreateCompilation(
				ScriptParameters parameters,
				SourceText sourceText,
				IEnumerable<MetadataReference> references)
			{
				SyntaxTree syntaxTree = ParseScript(parameters, sourceText);
				VisualBasicCompilationOptions options = CreateCompilationOptions(parameters);
				VisualBasicCompilation result = VisualBasicCompilation.Create(parameters.ScriptName, new[] { syntaxTree }, references, options);
				return result;
			}

			#endregion

			#region Private Methods

			private static VisualBasicParseOptions CreateParseOptions(ScriptParameters parameters)
			{
				Dictionary<string, object> symbols = new Dictionary<string, object>();
				symbols.Add("TRACE", true);
				if (parameters.Debug)
				{
					symbols.Add("DEBUG", true);
				}

				VisualBasicParseOptions parseOptions = new VisualBasicParseOptions(
					documentationMode: DocumentationMode.None,
					kind: SourceCodeKind.Regular,
					preprocessorSymbols: symbols);

				return parseOptions;
			}

			private static SyntaxTree ParseScript(ScriptParameters parameters, SourceText sourceText)
			{
				VisualBasicParseOptions parseOptions = CreateParseOptions(parameters);

				// We must tell the parser the FileName that the sourceText came from in order for debugging
				// to work (e.g., to have file and line info in the .pdb).
				SyntaxTree result = VisualBasicSyntaxTree.ParseText(sourceText, parseOptions, parameters.FileName);
				return result;
			}

			private static VisualBasicCompilationOptions CreateCompilationOptions(ScriptParameters parameters)
			{
				IEnumerable<GlobalImport> globalImports = GlobalImport.Parse(ScriptVbProvider.GlobalImports);

				VisualBasicCompilationOptions result = new VisualBasicCompilationOptions(
					OutputKind.WindowsApplication,
					globalImports: globalImports,
					optionStrict: OptionStrict.On,
					optionInfer: true,
					optionExplicit: true,
					parseOptions: CreateParseOptions(parameters),
					checkOverflow: parameters.Debug);

				result = AddCommonOptions(result, parameters);
				return result;
			}

			#endregion
		}

		#endregion
	}
}
